import argparse
import json
import sys

from llama_cpp import Llama


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--input-column")
    parser.add_argument("--output-column")
    parser.add_argument("--prompt")
    parser.add_argument("--model-path")
    parser.add_argument("--max-tokens", type=int)
    parser.add_argument("--echo", action="store_true")

    args = parser.parse_args()

    llm = Llama(model_path=args.model_path)

    for line in sys.stdin:
        data = json.loads(line)
        input_row = str(data[args.input_column])
        prepared_prompt = args.prompt.replace("{{value}}", input_row)
        processed_row = llm.create_completion(
            prepared_prompt, max_tokens=args.max_tokens, echo=args.echo,
        )["choices"][0]["text"]
        data[args.output_column] = processed_row
        sys.stdout.write(json.dumps(data))


if __name__ == "__main__":
    main()



import abc

import attr

from infernosaurus import models


@attr.define
class OnlineInferenceBackendBase(abc.ABC):
    runtime_config: models.OnlineInferenceRuntimeConfig = attr.ib()

    @abc.abstractmethod
    def get_operation_spec(self):
        pass

    @abc.abstractmethod
    def is_ready(self, runtime_info: models.OnlineInferenceRuntimeInfo) -> bool:
        pass


@attr.define
class OfflineInferenceBackendBase(abc.ABC):
    runtime_config: models.OfflineInferenceRuntimeConfig = attr.ib()

    @abc.abstractmethod
    def get_operation_spec(self, request: models.OfflineInferenceRequest):
        pass


import os

import pytest
import yt.wrapper as yt
from testcontainers_yt_local.container import YtLocalContainer


def _get_data_path(rel_path: str):
    return os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "data",
        rel_path,
    )


@pytest.fixture(scope="session")
def yt_with_model():
    yt_container = YtLocalContainer(
        use_ng_image=True, enable_cri_jobs=True, privileged=True,
    )
    for port in range(24578, 24578 + 4):
        yt_container = yt_container.with_bind_ports(port, port)

    with yt_container:
        yt_cli: yt.YtClient = yt_container.get_client(token="topsecret")
        yt_cli.write_file("//tmp/the-model.gguf", open(_get_data_path("qwen-494M-0.5-F16.gguf"), "rb"))

        yield yt_container


import yt.wrapper as yt

from infernosaurus.inference_operator import OnlineInferenceOperator, OfflineInferenceOperator
from infernosaurus import const as c
from infernosaurus import models


CLIENT_CONFIG_PATCH = {"is_local_mode": True, "proxy": {"enable_proxy_discovery": False}}


def test_start_and_stop(yt_with_model):
    llm = OnlineInferenceOperator(
        runtime_config=models.OnlineInferenceRuntimeConfig(
            yt_settings=models.YtSettings(
                proxy_url=yt_with_model.proxy_url_http, token="topsecret",
                client_config_patch=CLIENT_CONFIG_PATCH,
            ),
            server_resources=models.Resources(mem=10 * c.GiB, cpu=1),
            model_path="//tmp/the-model.gguf",
            operation_title="llama's ass"
        ),
        backend_type="llama_cpp",
    )
    try:
        llm.start()

        yt_cli: yt.YtClient = yt_with_model.get_client(token="topsecret")

        ops = yt_cli.list_operations(state="running")["operations"]
        assert len(ops) == 1
        op = ops[0]
        assert op["brief_spec"]["title"] == "llama's ass"
    finally:
        try:
            llm.stop()
        except Exception:
            raise


def test_server_only(yt_with_model):
    with OnlineInferenceOperator(
        backend_type="llama_cpp",
        runtime_config=models.OnlineInferenceRuntimeConfig(
            yt_settings=models.YtSettings(
                proxy_url=yt_with_model.proxy_url_http, token="topsecret",
                client_config_patch=CLIENT_CONFIG_PATCH,
            ),
            server_resources=models.Resources(mem=10 * c.GiB, cpu=1),
            model_path="//tmp/the-model.gguf",
        )
    ) as llm:
        openai_client = llm.get_openai_client()

        chat_completion = openai_client.chat.completions.create(
            messages=[
                {
                    "role": "user",
                    "content": "Write a recipe of an apple pie",
                }
            ],
            model="the-model.gguf",
        )
        content = chat_completion.choices[0].message.content

        chat_completion = openai_client.chat.completions.create(
            messages=[
                {
                    "role": "user",
                    "content": f"What is the following text about?\n{content}",
                }
            ],
            model="the-model.gguf",
        )
        content = chat_completion.choices[0].message.content

        assert "apple pie" in content.lower()


def test_with_workers(yt_with_model):
    with OnlineInferenceOperator(
        backend_type="llama_cpp",
        runtime_config=models.OnlineInferenceRuntimeConfig(
            yt_settings=models.YtSettings(
                proxy_url=yt_with_model.proxy_url_http, token="topsecret",
                client_config_patch=CLIENT_CONFIG_PATCH,
            ),
            server_resources=models.Resources(mem=3 * c.GiB, cpu=1),
            worker_num=3,
            worker_resources=models.Resources(mem=3 * c.GiB, cpu=1),
            model_path="//tmp/the-model.gguf",
        )
    ) as llm:
        openai_client = llm.get_openai_client()

        chat_completion = openai_client.chat.completions.create(
            messages=[
                {
                    "role": "user",
                    "content": "What is the capital of the Netherlands?",
                }
            ],
            model="the-model.gguf",
        )
        content = chat_completion.choices[0].message.content
        assert "Amsterdam" in content


def test_offline(yt_with_model):
    yt_cli: yt.YtClient = yt_with_model.get_client(token="topsecret")
    yt_cli.create("table", "//tmp/my_table")
    yt_cli.write_table(
        "//tmp/my_table",
        [
            {"number": "one", "country": "Germany", "true_answer": "Berlin"},
            {"number": "two", "country": "Italy", "true_answer": "Rome"},
            {"number": "three", "country": "Spain", "true_answer": "Madrid"},
            {"number": "four", "country": "France", "true_answer": "Paris"},
            {"number": "five", "country": "Armenia", "true_answer": "Yerevan"},
            {"number": "six", "country": "Serbia", "true_answer": "Belgrade"},
        ]
    )

    llm = OfflineInferenceOperator(
        backend_type="llama_cpp",
        runtime_config=models.OfflineInferenceRuntimeConfig(
            yt_settings=models.YtSettings(
                proxy_url=yt_with_model.proxy_url_http, token="topsecret",
                client_config_patch=CLIENT_CONFIG_PATCH,
            ),
            worker_num=2,
            worker_resources=models.Resources(cpu=4, mem=8 * c.GiB),
        )
    )
    llm.process(models.OfflineInferenceRequest(
        input_table="//tmp/my_table", input_column="country",
        output_table="//tmp/new_table", output_column="answer",
        prompt="Question: What is the capital of {{value}}? Answer:",
        model_path="//tmp/the-model.gguf", echo=True,
    ))

    data = list(yt_cli.read_table("//tmp/new_table"))

    for idx, row in enumerate(data):
        assert row["true_answer"] in row["answer"], row



import time
from typing import Any

import openai
import yt.wrapper as yt

from infernosaurus.backends.llama_cpp.backend import LlamaCppOnline, LlamaCppOffline
from infernosaurus.inference_backend_base import OnlineInferenceBackendBase, OfflineInferenceBackendBase
from infernosaurus.models import (
    OnlineInferenceRuntimeInfo,
    OnlineInferenceRuntimeConfig,
    OfflineInferenceRuntimeConfig,
    OfflineInferenceRequest,
)


class OnlineInferenceOperator:
    _backend: OnlineInferenceBackendBase
    _runtime_info: OnlineInferenceRuntimeInfo | None

    def __init__(self, backend_type: str, runtime_config: OnlineInferenceRuntimeConfig):
        self.yt_client = yt.YtClient(
            proxy=runtime_config.yt_settings.proxy_url, token=runtime_config.yt_settings.token,
            config=runtime_config.yt_settings.client_config_patch,
        )

        self._backend = {
            "llama_cpp": LlamaCppOnline,
        }[backend_type](runtime_config=runtime_config)

    def _get_job_hostport(self, job: dict[str, Any]) -> tuple[str, int]:
        exec_node_address = job["address"]
        job_id = job["id"]
        ports = self.yt_client.get(
            f"//sys/exec_nodes/{exec_node_address}/orchid/exec_node/job_controller/active_jobs/{job_id}/job_ports"
        )
        port = ports[0]
        host = exec_node_address.split(":")[0]
        return host, port

    def start(self):
        op_spec = self._backend.get_operation_spec()
        op = self.yt_client.run_operation(op_spec, sync=False)

        # WAIT FOR NUM JOBS
        for i in range(300):
            op_jobs = self.yt_client.list_jobs(op.id)["jobs"]
            if len(op_jobs) == self._backend.runtime_config.worker_num + 1:  # FIXME
                break

            if i % 10 == 0:
                op_state = self.yt_client.get_operation_state(op.id)
                if op_state == "failed":
                    # TODO: exception class
                    # TODO: stderr
                    raise Exception("Operation failed")
            time.sleep(1)

        server_job = next(j for j in op_jobs if j["task_name"] == "server")
        server_host, server_port = self._get_job_hostport(server_job)

        # FILL IN RUNTIME INFO
        self._runtime_info = OnlineInferenceRuntimeInfo(
            operation_id=op.id,
            server_job_id=server_job["id"],
            server_url=f"http://{server_host}:{server_port}",  # TODO: https
        )

        # check server is ready
        while True:
            if not self._backend.is_ready(self._runtime_info):
                time.sleep(1)
            else:
                break

    def stop(self) -> None:
        self.yt_client.abort_operation(self._runtime_info.operation_id)
        self._runtime_info = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def get_openai_client(self) -> openai.OpenAI:
        return openai.OpenAI(
            # TODO: https
            base_url=self._runtime_info.server_url,
            api_key="no-key",
        )


class OfflineInferenceOperator:
    _backend: OfflineInferenceBackendBase

    def __init__(self, backend_type: str, runtime_config: OfflineInferenceRuntimeConfig):
        self.yt_client = yt.YtClient(
            proxy=runtime_config.yt_settings.proxy_url, token=runtime_config.yt_settings.token,
            config=runtime_config.yt_settings.client_config_patch,
        )

        self._backend = {
            "llama_cpp": LlamaCppOffline,
        }[backend_type](runtime_config=runtime_config)

    def process(self, request: OfflineInferenceRequest):
        op_spec = self._backend.get_operation_spec(request)
        self.yt_client.run_operation(op_spec)


from typing import Any

import attr

from infernosaurus import typing as t


@attr.define
class OnlineInferenceRuntimeInfo:
    operation_id: t.OpID = attr.ib()
    server_job_id: t.JobID = attr.ib()
    server_url: str = attr.ib()


@attr.define
class Resources:
    cpu: int = attr.ib()
    mem: int = attr.ib()


@attr.define
class YtSettings:
    proxy_url: str = attr.ib()
    token: str = attr.ib(repr=False)
    client_config_patch: dict[str, Any] = attr.ib(factory=dict)


@attr.define
class OnlineInferenceRuntimeConfig:
    yt_settings: YtSettings = attr.ib()
    server_resources: Resources = attr.ib()
    model_path: str = attr.ib()
    worker_num: int = attr.ib(default=0)
    worker_resources: Resources | None = attr.ib(default=None)
    operation_title: str = attr.ib(default="llama's ass")


@attr.define
class OfflineInferenceRuntimeConfig:
    yt_settings: YtSettings = attr.ib()
    worker_num: int = attr.ib()
    worker_resources: Resources = attr.ib()


@attr.define
class OfflineInferenceRequest:
    input_table: str = attr.ib()
    input_column: str = attr.ib()
    output_table: str = attr.ib()
    output_column: str = attr.ib()
    model_path: str = attr.ib()
    prompt: str = attr.ib()
    echo: bool = attr.ib(default=False)
    max_tokens: int = attr.ib(default=256)


import argparse
import os
import subprocess
import sys
import time

import yt.wrapper as yt


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-workers", type=int)
    parser.add_argument("--model")
    args = parser.parse_args()

    yt_client = yt.YtClient(proxy=os.environ["YT_PROXY"], token=os.environ["YT_SECURE_VAULT_YT_TOKEN"])
    yt_client.config["proxy"]["url"] = "http://localhost:80"  # FIXME

    operation_id = os.environ["YT_OPERATION_ID"]

    while True:
        op_jobs = yt_client.list_jobs(operation_id)["jobs"]
        if len(op_jobs) == args.num_workers + 1:
            break
        time.sleep(1)

    worker_jobs = [j for j in op_jobs if j["task_name"] == "workers"]
    assert len(worker_jobs) == args.num_workers, len(worker_jobs)

    worker_addresses = []
    for wjob in worker_jobs:
        exec_node_address = wjob["address"]
        job_id = wjob["id"]
        ports = yt_client.get(f"//sys/exec_nodes/{exec_node_address}/orchid/exec_node/job_controller/active_jobs/{job_id}/job_ports")
        port = ports[0]
        host = exec_node_address.split(":")[0]
        worker_addresses.append(f"{host}:{port}")

    run_server_command = [
        "/llama/bin/llama-server", "-m", args.model,
        "-ngl", "99", "--host", "0.0.0.0", "--port", os.environ["YT_PORT_0"],
    ]
    if len(worker_addresses) > 0:
        run_server_command.extend(["--rpc", ",".join(worker_addresses)])

    subprocess.run(run_server_command, stdout=sys.stderr, check=True)


if __name__ == "__main__":
    main()



import os.path

import httpx
import yt.wrapper as yt

from infernosaurus.inference_backend_base import OnlineInferenceBackendBase, OfflineInferenceBackendBase
from infernosaurus import models
from infernosaurus.utils import quoted as q


class LlamaCppOffline(OfflineInferenceBackendBase):
    def get_operation_spec(self, request: models.OfflineInferenceRequest):
        infer_script_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "infer.py",
        )
        model_rel_path = "./" + request.model_path.split("/")[-1]

        command = " ".join([
            "python3", "./infer.py", "--model-path", q(model_rel_path),
            "--input-column", q(request.input_column), "--output-column", q(request.output_column),
            "--prompt", q(request.prompt), "--echo" if request.echo else "",
            "--max-tokens", str(request.max_tokens),
        ])

        op_spec = yt.MapSpecBuilder() \
            .begin_mapper() \
                .command(command) \
                .format(yt.JsonFormat(encode_utf8=False)) \
                .docker_image("ghcr.io/dmi-feo/llamosaurus:4") \
                .memory_limit(self.runtime_config.worker_resources.mem) \
                .cpu_limit(self.runtime_config.worker_resources.cpu) \
                .file_paths([request.model_path, yt.LocalFile(infer_script_path)]) \
            .end_mapper() \
            .input_table_paths([request.input_table]) \
            .output_table_paths([request.output_table]) \
            .job_count(self.runtime_config.worker_num) \
            .stderr_table_path("//tmp/stderr") \
            .max_failed_job_count(1)

        return op_spec


class LlamaCppOnline(OnlineInferenceBackendBase):
    def get_operation_spec(self):  # TODO: typing
        op_spec = yt.VanillaSpecBuilder()
        op_spec = self._build_server_task(op_spec)
        if self.runtime_config.worker_num > 0:
            op_spec = self._build_workers_task(op_spec)

        op_spec = op_spec \
            .stderr_table_path("//tmp/stderr") \
            .max_failed_job_count(1) \
            .secure_vault({"YT_TOKEN": self.runtime_config.yt_settings.token}) \
            .title(self.runtime_config.operation_title)

        return op_spec

    def is_ready(self, runtime_info: models.OnlineInferenceRuntimeInfo) -> bool:
        try:
            resp = httpx.get(f"{runtime_info.server_url}/health")
        except (httpx.NetworkError, httpx.ProtocolError):
            return False
        return resp.status_code == 200

    def _build_server_task(self, op_spec_builder):
        bootstrap_script_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "bootstrap_server.py",
        )
        model_rel_path = "./" + self.runtime_config.model_path.split("/")[-1]

        return op_spec_builder.begin_task("server") \
            .command(f"python3 ./bootstrap_server.py --num-workers {self.runtime_config.worker_num} --model {model_rel_path}") \
            .job_count(1) \
            .docker_image("ghcr.io/dmi-feo/llamosaurus:2") \
            .port_count(1) \
            .memory_limit(self.runtime_config.server_resources.mem) \
            .cpu_limit(self.runtime_config.server_resources.cpu) \
            .environment({"YT_ALLOW_HTTP_REQUESTS_TO_YT_FROM_JOB": "1", "YT_PROXY": self.runtime_config.yt_settings.proxy_url}) \
            .file_paths([self.runtime_config.model_path, yt.LocalFile(bootstrap_script_path)]) \
            .end_task()

    def _build_workers_task(self, op_spec_builder):
        return op_spec_builder.begin_task("workers") \
            .command("/llama/bin/rpc-server --host 0.0.0.0 --port $YT_PORT_0 >&2") \
            .job_count(self.runtime_config.worker_num) \
            .docker_image("ghcr.io/dmi-feo/llamosaurus:2") \
            .port_count(1) \
            .memory_limit(self.runtime_config.worker_resources.mem) \
            .cpu_limit(self.runtime_config.worker_resources.cpu) \
            .end_task()